@@ -19,12 +19,12 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-VERSION = '1.20120210'
+VERSION = '1.20121210'
 
 DESCRIPTION = '''\
 go2 is a fast directory finder.
 
-This is version {0}, Copyright (C) 2004-2011 David Villa Alises.
+This is version {0}, Copyright (C) 2004-2012 David Villa Alises.
 go2 comes with ABSOLUTELY NO WARRANTY; This is free software, and you
 are welcome to redistribute it under certain conditions; See COPYING
 for details.'''
@@ -49,6 +49,7 @@ import signal
 import re
 import logging
 import locale
+import atexit
 
 from threading import Thread
 import multiprocessing as mp
@@ -122,27 +123,29 @@ class PathItem:
 
 
 def go2setup():
+    bashrc = os.path.join(USERDIR, '.bashrc')
+    open_flags = 'a'
+
     try:
-        if ACTIVATION_CMD in config.fs.getcontents(
-            os.path.join(USERDIR, '.bashrc')):
+        if ACTIVATION_CMD in config.fs.getcontents(bashrc):
             print('go2 already configured, skipping.')
             return 1
 
     except ResourceNotFoundError:
-        pass
+        open_flags = 'w'
 
-    with config.fs.open(os.path.join(USERDIR, '.bashrc'), 'a') as fd:
+    with config.fs.open(os.path.join(USERDIR, '.bashrc'), open_flags) as fd:
         fd.write("{0}\n{1}\n".format(ACTIVATION_CMD, CD_MODE_CMD))
         print('Setting up go2. It will be ready in next shell session.')
 
 
 def rprint(text=''):
     sys.stdout.write("\r{0}".format(text.encode(config.encoding)))
-    sys.stdout.flush()
+#    sys.stdout.flush()
 
 
 def message(text):
-    rprint("({0})\n".format(text))
+    rprint("({0})\r\n".format(text))
 
 
 def high(text):
@@ -344,20 +347,10 @@ class CommandProvider:
         self.abort = False
         signal.signal(signal.SIGTERM, self.terminate)
 
-#    def is_alive2(self):
-#        try:
-#            self.ps.send_signal(0)
-#        except OSError:
-#            return False
-#
-#        return True
-
     def is_alive(self):
-#        return self.ps.poll() is None
         value = os.waitpid(self.ps.pid, os.WNOHANG)
         logging.debug('%s: waitpid %s ', self, value)
         return value == (0, 0)
-#        return os.waitpid(self.ps.pid, os.WNOHANG) == (0, 0)
 
     def terminate(self, *args):
         if self.abort:
@@ -408,14 +401,33 @@ class CommandProvider:
 
 
 def tree_provider(path):
-    path = os.path.abspath(path)
-
-    if not config.fs.isdir(path):
-        message("'{0}' does not exist".format(path))
+    assert os.path.isabs(path)
+    assert config.fs.isdir(path)
 
     return CommandProvider('tree -dfin --noreport %s' % path)
 
 
+def walk_provider(path):
+    print 'walk_provider'
+    assert os.path.isabs(path)
+    assert config.fs.isdir(path)
+
+    for root, dirnames, filenames in os.walk(path):
+        paths = [os.path.join(root, x) for x in dirnames
+                 if not x.startswith('.')]
+
+        for path in paths:
+            #FIXME: refactor this, is repeated in all providers
+            level = config.matcher.match(path)
+            if level == MatchLevel.NO_MATCH:
+                continue
+
+            if not config.fs.exists(path):
+                continue
+
+            yield PathItem(path, level)
+
+
 class CancelException(Go2Exception):
     pass
 
@@ -507,12 +519,6 @@ class ProcessPool(object):
 
         time.sleep(0.1)
 
-#        term_threads = [ThreadStarted(target=w.terminate) for w in self.workers
-#                        if w.is_alive()]
-#
-#        for t in term_threads:
-#            t.join()
-
         for w in self.workers:
             if w.is_alive():
                 w.terminate()
@@ -628,7 +634,7 @@ class PathBuffer(object):
 #        print '\r', self.groups[item.level]
 
         if item.path in self.groups[item.level] or \
-           any(f(item.path) for f in self.filters):
+                any(f(item.path) for f in self.filters):
             return
 
         self.groups[item.level].append(item.path)
@@ -693,7 +699,7 @@ class Menu(Sink):
             raise self.InvalidChoiceException()
 
         logging.debug("choice: {0}, entries: {1}".format(
-                choice, self.entries))
+                      choice, self.entries))
 
         self.target = self.entries[choice]
         self.reactor.quit()
@@ -800,13 +806,18 @@ class Go2Base(object):
         self.reactor = QueueReactor()
         self.path_buffer = PathBuffer()
 
-        self.setup_ignore()
-        self.setup_tty()
+        self.setup_stdin()
 
+        self.setup_ignore()
         self.reactor.queue_add_watch(self.pool.output_queue,
                                      QueueExtractor(self.path_buffer.add))
         self.reactor.timeout_add(250, self.end_checker)
 
+    def setup_stdin(self):
+        stdin_tty = TTY()
+        stdin_tty.set_raw()
+        atexit.register(stdin_tty.restore)
+
     def setup_ignore(self):
         try:
             ignore_manager = IgnoreManager.from_file(GO2IGNORE)
@@ -814,10 +825,8 @@ class Go2Base(object):
         except ResourceNotFoundError:
             pass
 
-    def setup_tty(self):
-        tty = TTY()
-        tty.set_raw()
-        self.reactor.at_quit(tty.restore)
+    def stop(self, *args):
+        self.pool.terminate()
 
     def run(self):
         retval = 1
@@ -848,12 +857,28 @@ class Go2Base(object):
         self.pool.sleep_free_workers(seconds=0.5)
         self.pool.add_task(from_file_provider, self.cache)
 
+    def not_overlapped_history(self):
+        history_paths = []
+        for p in sorted(self.history.data.keys(), key=len):
+            if any(p.startswith(x) for x in history_paths):
+                continue
+
+            history_paths.append(p)
+
+#        print str.join('\r\n', history_paths)
+        return history_paths
+
     def create_command_tasks(self):
         paths = ListUniq(config.search_path.split(':'))
 #        paths.extend(self.not_overlapped_history())
 
-        for path in paths:
+        for path in set(paths):
+            path = os.path.abspath(path)
+            if not config.fs.isdir(path):
+                message("'{0}' does not exist".format(path))
+
             self.pool.add_task(tree_provider, path)
+#            self.pool.add_task(walk_provider, path)
 
     def end_checker(self):
         return False
@@ -865,6 +890,8 @@ class Go2Base(object):
 class Go2Interactive(Go2Base):
     def __init__(self):
         Go2Base.__init__(self)
+        self.setup_stdout()
+
         self.menu = Menu(self.reactor, sys.stdout)
         self.menu.formatter = high_pattern_formatter
         self.path_buffer.set_sink(self.menu)
@@ -873,6 +900,17 @@ class Go2Interactive(Go2Base):
 
         self.progress = cycle(range(4))
 
+    def setup_stdout(self):
+        def stdout_restore():
+            sys.stdout = sys.__stdout__
+
+        sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)
+        atexit.register(stdout_restore)
+
+# fl = fcntl.fcntl(fd.fileno(), fcntl.F_GETFL)
+# fl |= os.O_SYNC # or os.O_DSYNC (if you don't care the file timestamp updates)
+# fcntl.fcntl(fd.fileno(), fcntl.F_SETFL, fl)
+
     def end_checker(self):
         if self.menu.is_full:
             message(u"warning: too many matches!")
@@ -898,8 +936,8 @@ class Go2Interactive(Go2Base):
         return False
 
     def on_success(self):
-        rprint(u"Changing to: {0}\n".format(
-                high(abbreviate_home_prefix(self.menu.target))))
+        rprint(u"Changing to: {0}\r\n".format(
+               high(abbreviate_home_prefix(self.menu.target))))
 
         save_target(self.menu.target)
 
@@ -925,13 +963,15 @@ class Go2ListOnly(Go2Base):
 def create_directory_wizzard(path):
 
     def show_make_and_change(path):
+        path = os.path.abspath(path)
         path = config.matcher.high_pattern(path)
         path = abbreviate_home_prefix(path)
-        print("go2: Making and changing to directory:\n%s." % path)
+        print("go2: Making and changing to directory: %s" % path)
 
     if path.startswith(CWD):
-        target = os.path.abspath(path)[len(CWD):].strip('/')
-    print "go2: '%s' does not exist.\n(c)ancel, (s)earch or (m)ake? (C/s/m):" % target,
+        path = os.path.abspath(path)[len(CWD):].strip('/')
+
+    print "go2: '%s' does not exist.\n(c)ancel, (s)earch or (m)ake? (C/s/m):" % path,
 
     try:
         answer = raw_input().lower()
@@ -971,6 +1011,10 @@ def chdir_decorator(wizzard=False):
         logging.error(e)
         save_target(target)
         return 0
+    except OSError, e:
+        logging.error(e)
+        save_target(USERDIR)
+        return 0
 
     if params == '-':
         save_target('-')
@@ -1062,5 +1106,5 @@ if __name__ == '__main__':
         config.parser.print_help()
         sys.exit(1)
 
-    signal.signal(signal.SIGINT, signal.SIG_IGN)
+    signal.signal(signal.SIGINT, config.engine.stop)
     sys.exit(config.engine().run())
