@@ -1,9 +1,9 @@
-#!/bin/bash
+#!/bin/sh --
 GO2TMP=~/.go2/tmp
-function go2 {
+go2 () {
     /usr/lib/go2/go2.py $*;
     if [ -e $GO2TMP ]; then
-	    cd "`cat $GO2TMP`"
-	    rm $GO2TMP > /dev/null
+	cd "$(cat $GO2TMP)"
+	rm $GO2TMP > /dev/null
     fi
 }
